/*============================================================================
//
// Copyright(c) 2006 Intel Corporation. All rights reserved.
//   All rights reserved.
// 
//   Redistribution and use in source and binary forms, with or without 
//   modification, are permitted provided that the following conditions 
//   are met:
// 
//     * Redistributions of source code must retain the above copyright 
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright 
//       notice, this list of conditions and the following disclaimer in 
//       the documentation and/or other materials provided with the 
//       distribution.
//     * Neither the name of Intel Corporation nor the names of its 
//       contributors may be used to endorse or promote products derived 
//       from this software without specific prior written permission.
// 
//   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
//   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
//   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
//   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
//   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
//   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
//   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
//   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//  File Name: RegistrarSM.cpp
//  Description: Implements Registrar State Machine Class.
//
****************************************************************************/

// RegistrarSM.cpp
#ifdef WIN32
#include <windows.h>
#endif

#include <string.h>

//OpenSSL includes
#include <openssl/bn.h>
#include <openssl/dh.h>

#include "WscHeaders.h"
#include "WscCommon.h"
#include "WscError.h"
#include "Portability.h"
#include "WscQueue.h"
#include "tutrace.h"
//#include "slist.h"
//#include "OobUfd.h"
#include "Transport.h"
#include "StateMachineInfo.h"
#include "RegProtoMsgs.h"
#include "RegProtocol.h"
#include "StateMachine.h"

// Include the UPnP Microsostack header file so an AP can send an EAP_FAIL right after sending M2D if there
// are no currently-subscribed external Registrars.
#include "UPnPMicroStack.h" 

bool b_UpnpDevGetDeviceInfo = false;
CRegistrarSM::CRegistrarSM(IN CRegProtocol *pc_regProt, IN CTransport *pc_trans)
        :CStateMachine(pc_regProt, pc_trans, MODE_REGISTRAR), m_sentM2(false)
{
	m_ptimerStatus = 0;
	m_stophandlemsg = 0;
    WscSyncCreate(&mp_ptimerLock);
}

CRegistrarSM::~CRegistrarSM()
{
    WscSyncDestroy(mp_ptimerLock);
}

uint32 
CRegistrarSM::InitializeSM(IN S_DEVICE_INFO *p_enrolleeInfo, 
                           bool enableLocalSM,
                           bool enablePassthru)
{
    if((!enableLocalSM) && (!enablePassthru))
    {
        TUTRACE((TUTRACE_ERR, "REGSM: enableLocalSM and enablePassthru are "
                              "both false.\n"));
        return WSC_ERR_INVALID_PARAMETERS;
    }

    //if the device is an AP, passthru MUST be enabled
    if((mps_localInfo->b_ap) && (!enablePassthru))
    {
        TUTRACE((TUTRACE_ERR, "REGSM: AP Passthru not enabled.\n"));
        return WSC_ERR_INVALID_PARAMETERS;
    }

    m_localSMEnabled = enableLocalSM;
    m_passThruEnabled = enablePassthru;

    uint32 err = CStateMachine::InitializeSM();
    if(WSC_SUCCESS != err)
        return err;

    mps_regData->p_enrolleeInfo = p_enrolleeInfo;
    mps_regData->p_registrarInfo = mps_localInfo;
    m_protocoltimerThrdId = 0; 
    return WSC_SUCCESS;
}

//This method is used by the Proxy to send packets to 
//the various registrars. If the local registrar is 
//enabled, this method will call it using RegStepSM
uint8 real_mac[6] = {0x00,0x90,0x4C, 0x68, 0x08, 0xB9};

void set_realmac(uint8 *mac)
{
	memcpy(real_mac, mac, 6);
	return;
}

uint8 *get_realmac()
{
	return real_mac;
}
uint32 
CRegistrarSM::Step(IN uint32 msgLen, IN uint8 *p_msg)
{
    uint32 err2,err = WSC_SUCCESS;
				
    TUTRACE((TUTRACE_INFO, "REGSM: Entering Step.\n"));
    if(false == m_initialized)
    {
        TUTRACE((TUTRACE_ERR, "REGSM: Not yet initialized.\n"));
        return WSC_ERR_NOT_INITIALIZED;
    }

    //Irrespective of whether the local registrar is enabled or whether we're 
    //using an external registrar, we need to send a WSC_Start over EAP to
    //kickstart the protocol.
    //If we send a WSC_START msg, we don't need to do anything else i.e. 
    //invoke the local registrar or pass the message over UPnP
    if( (!msgLen) && (START == mps_regData->e_smState)&& 
        (  (TRANSPORT_TYPE_EAP == m_transportType) ||
           (TRANSPORT_TYPE_UPNP_CP == m_transportType)) )

    {
        err = mpc_trans->SendStartMessage(m_transportType);
        if(WSC_SUCCESS != err)
        {
            TUTRACE((TUTRACE_ERR, "REGSM: SendStartMessage failed. Err = %d\n",
                     err));
        }
        return err; //no more processing for WSC_START
    }
    
    //Now send the message over UPnP. Else, if the message has come over UPnP
    //send it over EAP.
    //if(m_passThruEnabled)

    //temporarily disable forwarding to/from UPnP if we've already sent M2
    //this is a stop-gap measure to prevent multiple registrars from cluttering
    //up the EAP session.
    if((m_passThruEnabled) && (!m_sentM2))
    {
        switch(m_transportType)
        {
            case TRANSPORT_TYPE_EAP:
            {
                TUTRACE((TUTRACE_INFO, "REGSM: Forwarding message from "
                                        "EAP to UPnP.\n"));
                TRANSPORT_TYPE trType;
            //get correct MAC address from client
            if (msgLen) 
            {
                uint32 msgType2;
                uint8 *macaddr;
                BufferObj *inMsg = NULL;
                BufferObj regProtoMsg2(p_msg, msgLen);
                inMsg = &regProtoMsg2;
                err = mpc_regProt->GetMsgType(msgType2, *inMsg);
                if(WSC_SUCCESS != err)
                {
                    TUTRACE((TUTRACE_INFO, "REGSM: Get message type error. \n"));
                }
                else
                {
                    if(WSC_ID_MESSAGE_M1 == msgType2)
                    {
                        TUTRACE((TUTRACE_ERR, "REGSM: M1 message got!\n"));
                	    S_WSC_M1 m1;
                
                        m1.version       = CTlvVersion(WSC_ID_VERSION, *inMsg);
                        m1.msgType       = CTlvMsgType(WSC_ID_MSG_TYPE, *inMsg);
                        m1.uuid          = CTlvUuid(WSC_ID_UUID_E, *inMsg, SIZE_UUID);
                        m1.macAddr       = CTlvMacAddr(WSC_ID_MAC_ADDR, *inMsg, SIZE_MAC_ADDR);
                		macaddr = m1.macAddr.Value();
                
                        TUTRACE((TUTRACE_ERR, "REGSM: MAC in message M1 is: %02x:%02x:%02x:%02x:%02x:%02x\n",
                                  macaddr[0], macaddr[1], macaddr[2], macaddr[3], macaddr[4], macaddr[5]));
                
                		set_realmac(macaddr);	
                    }
					else
                        TUTRACE((TUTRACE_INFO,"\n###EAP to UPnP msgType=%x ###\n",msgType2));

                }
            }

                if(mps_localInfo->b_ap)
                    trType = TRANSPORT_TYPE_UPNP_DEV;
                else
                    trType = TRANSPORT_TYPE_UPNP_CP;

				BufferObj regProtoMsg(p_msg, msgLen);
				uint32 msgType;

				err = mpc_regProt->GetMsgType(msgType, regProtoMsg);
                //recvd over EAP, send over UPnP
                if (msgType != WSC_ID_MESSAGE_ACK) //borg
                mpc_trans->TrWrite(trType, 
                                    (char *)p_msg, 
                                    msgLen);
				// Now check to see if the EAP message is a WSC_NACK.  If so, 
				// terminate the EAP connection with an EAP-Fail.  

				if(WSC_SUCCESS == err && msgType == WSC_ID_MESSAGE_NACK)
				{
					TUTRACE((TUTRACE_ERR, "REGSM: Notifying MC of failure.\n"));
					NotifyMasterControl(SM_FAILURE, NULL, NULL);

					//reset the transport connection
					mpc_trans->TrWrite(m_transportType, NULL, 0);

					//reset the SM
					RestartSM();
					TUTRACE((TUTRACE_ERR, "REGSM: TrWrite generated an "
											"error: %d\n", err));
					return WSC_MESSAGE_PROCESSING_ERROR;
				}

				if(WSC_SUCCESS == err && (msgType == WSC_ID_MESSAGE_DONE || msgType == WSC_ID_MESSAGE_ACK))
				{
					//borg Send EAP fail when DONE or ACK
					TUTRACE((TUTRACE_INFO, "REGSM: Send EAP fail when DONE or ACK (type=%x).\n",msgType));					
                    mpc_trans->TrWrite(TRANSPORT_TYPE_EAP, NULL, 0);
				}			
            }
            break;
            default:
            {
				uint32 msgType3;
				BufferObj regProtoMsg3(p_msg, msgLen);
                TUTRACE((TUTRACE_INFO, "REGSM: Forwarding message from "
                                        "UPnP to EAP.\n"));
                //recvd over UPNP, send it out over EAP
                TUTRACE((TUTRACE_INFO,"\n== GetDeviceInfo=%d ==\n",b_UpnpDevGetDeviceInfo));
				
                if (msgLen)
                {
                    err2 = mpc_regProt->GetMsgType(msgType3, regProtoMsg3);
		            TUTRACE((TUTRACE_INFO,"\n###UPnP to EAP msgType=%x ###\n",msgType3));
			
                    if(WSC_SUCCESS == err2 && msgType3 == WSC_ID_MESSAGE_M2D )
                    {
      		            TUTRACE((TUTRACE_INFO,"\n ====received M2D ignore when proxy mode!====\n"));
                    }
                }

                if(b_UpnpDevGetDeviceInfo)
                	GetDeviceInfoStep();	
                if(msgLen && msgType3 != WSC_ID_MESSAGE_M2D) //borg
                {
                    err = mpc_trans->TrWrite(TRANSPORT_TYPE_EAP, 
                        (char *)p_msg, 
                        msgLen);
                }
            }
        }//switch
    }

    //Send this message to the local registrar only if it hasn't
    //arrived from UPnP. This is needed on an AP with passthrough 
    //enabled so that the local registrar doesn't 
    //end up processing messages coming over UPnP from a remote registrar 
    if( (m_localSMEnabled) && (( ! m_passThruEnabled) ||
        ((m_transportType != TRANSPORT_TYPE_UPNP_DEV) && 
         (m_transportType != TRANSPORT_TYPE_UPNP_CP))))
    {
        TUTRACE((TUTRACE_INFO, "REGSM: Sending msg to local registrar.\n"));
        err = RegSMStep(msgLen, p_msg);
    }

    if(WSC_SUCCESS != err)
    {
        mps_regData->e_smState = FAILURE;
        TUTRACE((TUTRACE_ERR, "REGSM: TrWrite generated an "
                                "error: %d\n", err));
        return err;
    }

    return WSC_SUCCESS;
}


uint32
CRegistrarSM::GetDeviceInfoStep()
{
	uint32 err;
	//BufferObj *inMsg = NULL;
	char errMsg[256];
	BufferObj outBuf, tempBuf;

	b_UpnpDevGetDeviceInfo = false;
	err = mpc_regProt->BuildDeviceInfoRsp(mps_regData, tempBuf);
	if(WSC_SUCCESS != err)
	{
	    stringPrintf(errMsg, 256, "BuildMessageM1: %d", err);
	    throw errMsg;
	}
	//mps_regData->e_lastMsgSent = M1;
	outBuf.Append(tempBuf.Length(), tempBuf.GetBuf());
	//Now send the message to the transport
	err = mpc_trans->TrWrite(m_transportType,
                             (char *)outBuf.GetBuf(),
                             outBuf.Length());
	if(WSC_SUCCESS != err)
        {
	           mps_regData->e_smState = FAILURE;
              	TUTRACE((TUTRACE_ERR, "ENRSM: TrWrite generated an error: %d\n", err));
               	return WSC_ERR_GENERIC;
        }
}
uint32 
CRegistrarSM::RegSMStep(IN uint32 msgLen, IN uint8 *p_msg)

{
    uint32 err;
    BufferObj *inMsg = NULL;

    TUTRACE((TUTRACE_INFO, "REGSM: Entering RegSMStep.\n"));

    TUTRACE((TUTRACE_INFO, "REGSM: Recvd message of length %d\n", msgLen));

    if((!p_msg || !msgLen) && (START != mps_regData->e_smState))
    {
        TUTRACE((TUTRACE_ERR, "REGSM: Wrong input parameters.\n"));
        //Notify the MasterControl
        NotifyMasterControl(SM_FAILURE, NULL, NULL);
        return WSC_ERR_INVALID_PARAMETERS;
    }


    if((MNONE == mps_regData->e_lastMsgSent) && (msgLen))
    {
        //We've received a msg from the enrollee, we don't want to send
        //WSC_Start in this case
        //Change last message sent from MNONE to MSTART
        mps_regData->e_lastMsgSent = MSTART;
    }

    //S_WSC_HEADER * hdr = (S_WSC_HEADER *)p_msg;
    //if((hdr->opCode < WSC_Start) || (hdr->opCode > WSC_Done))
    //{
    //    TUTRACE((TUTRACE_ERR, "REGSM: Wrong opcode.\n"));
    //    //Notify the MasterControl
    //    m_initialized = false;
    //    NotifyMasterControl(SM_FAILURE, NULL, NULL);
    //    return SM_ERR_MESSAGE_DATA;
    //}

    //if(hdr->flags & 0x02)
    //{
    //    //the message length field is included
    //    //ignore it for now and just get the message
    //    temp = p_msg+sizeof(S_WSC_FRAGMENT_HEADER);
    //    tempLen = msgLen - sizeof(S_WSC_FRAGMENT_HEADER);
    //}
    //else
    //{
    //    temp = p_msg+sizeof(S_WSC_HEADER);
    //    tempLen = msgLen - sizeof(S_WSC_HEADER);
    //}

    //BufferObj regProtoMsg(temp, tempLen);
    //inMsg = &regProtoMsg;

    //switch(hdr->opCode)
    //{
    //case WSC_Start:
    //case WSC_ACK:
    //case WSC_NACK:
    //case WSC_Done:
    //case WSC_MSG:
    //    //First, check if we need to process this message.
    //    //If the nonce isn't ours, we should ignore it.
    //    if(tempLen)
    //    {
    //    }
    //        
    //    break;
    //default:
    //    //set the message state to failure
    //    mps_regData->e_smState = FAILURE;
    //}

    BufferObj regProtoMsg(p_msg, msgLen);
    inMsg = &regProtoMsg;

    err = mpc_regProt->CheckNonce(mps_regData->registrarNonce, 
                                    *inMsg, 
                                    WSC_ID_REGISTRAR_NONCE);
    //make an exception for M1. It won't have the registrar nonce
    if((WSC_SUCCESS != err) && (RPROT_ERR_REQD_TLV_MISSING != err))
    {
        TUTRACE((TUTRACE_INFO, "REGSM: Recvd message is not meant for"
                                " this registrar. Ignoring...\n"));
        return WSC_SUCCESS;
    }

    TUTRACE((TUTRACE_INFO, "REGSM: Calling HandleMessage...\n"));
    HandleMessage(*inMsg);

    //now check the state so we can act accordingly
    switch(mps_regData->e_smState)
    {
    case START:
    case CONTINUE:
        //do nothing.
        break;
    case RESTART:
        {
            //reset the SM
            RestartSM();
        }
        break;
    case SUCCESS:
        {
            TUTRACE((TUTRACE_ERR, "REGSM: Notifying MC of success.\n"));
            NotifyMasterControl(SM_SUCCESS, 
                                mps_regData->p_enrolleeInfo, 
                                mp_peerEncrSettings);
            
            //reset the transport connection
            mpc_trans->TrWrite(m_transportType, NULL, 0);

            //reset the SM
            RestartSM();
        }
        break;
    case FAILURE:
        {
            TUTRACE((TUTRACE_ERR, "REGSM: Notifying MC of failure.\n"));
            NotifyMasterControl(SM_FAILURE, NULL, NULL);

            //reset the transport connection
            mpc_trans->TrWrite(m_transportType, NULL, 0);

            //reset the SM
            RestartSM();
        }
        break;
    default:
        break;
    }
    return WSC_SUCCESS;
}

void CRegistrarSM::HandleMessage(BufferObj &msg)
{
    uint32 err;
    char errMsg[256];
    void *encrSettings = NULL;
    BufferObj tempBuf1;
    BufferObj tempBuf;
    BufferObj outBuf;
    uint32 msgType;
    
    try
    {
        //Append the header before doing any processing
        //S_WSC_HEADER hdr;
        //hdr.opCode = WSC_MSG;
        //hdr.flags = 0;
        //outBuf.Append(sizeof(hdr), (uint8 *)&hdr);

        err = mpc_regProt->GetMsgType(msgType, msg);
        if(WSC_SUCCESS != err)
        {
            stringPrintf(errMsg, 256, 
                            "ENRSM: GetMsgType returned error: %d\n", err);
            throw errMsg;
        }

		if (m_stophandlemsg == 1 )
        {
            //Now make the transport send an EAP fail
            TUTRACE((TUTRACE_INFO, "REGSM: Sending EAP FAIL\n"));
            mpc_trans->TrWrite(TRANSPORT_TYPE_EAP, NULL, 0);

            stringPrintf(errMsg, 256, 
                            "REGSM: Session timeout stop HandleMessage!!\n");
            throw errMsg;
        }			

        switch(mps_regData->e_lastMsgSent)
        {
        case MSTART:
            if(WSC_ID_MESSAGE_M1 != msgType)
            {
                TUTRACE((TUTRACE_ERR, "REGSM: Expected M1, received %d\n",
                          msgType));
                return;
            }

		    //Now, schedule a thread to sleep for some time to 
		    WscLock(mp_ptimerLock);
		    {
		        m_ptimerStatus = 0;
		        WscUnlock(mp_ptimerLock);
		        //return;
		    }

            err = mpc_regProt->ProcessMessageM1(mps_regData, msg);
            if(WSC_SUCCESS != err)
            {
                stringPrintf(errMsg, 256, "ProcessMessageM1: %d", err);
                return;//for vista //throw errMsg;
            }
            mps_regData->e_lastMsgRecd = M1;

            if(mps_regData->password.Length())
            {
                //We don't send any encrypted settings currently
                err = mpc_regProt->BuildMessageM2(mps_regData, tempBuf, NULL);
                if(WSC_SUCCESS != err)
                {
                    stringPrintf(errMsg, 256, "BuildMessageM2: %d", err);
					//Send a NACK to the Enrollee
					SendNack(WSC_MESSAGE_PROCESSING_ERROR);
                    throw errMsg;
                }
                mps_regData->e_lastMsgSent = M2;            
                //temporary change
                m_sentM2 = true;
                TUTRACE((TUTRACE_INFO, "REGSM: m_sentM2 = true !! %d\n"));				
            }
            else
            {
                //there is no password present, so send M2D
                err = mpc_regProt->BuildMessageM2D(mps_regData, tempBuf);
                if(WSC_SUCCESS != err)
                {
                    stringPrintf(errMsg, 256, "BuildMessageM2D: %d", err);
                    throw errMsg;
                }
                mps_regData->e_lastMsgSent = M2D;

                //Ask the master control to set the password
                //pass the enrollee's info to help the master control make
                //the appropriate decision
            }
            
            outBuf.Append(tempBuf.Length(), tempBuf.GetBuf());

            //Now send the message to the transport
            err = mpc_trans->TrWrite(m_transportType, 
                               (char *)outBuf.GetBuf(), 
                               outBuf.Length());
            if(WSC_SUCCESS != err)
            {
                mps_regData->e_smState = FAILURE;
                TUTRACE((TUTRACE_ERR, "REGSM: TrWrite generated an "
                                      "error: %d\n", err));
                return;
            }

            //Set state to CONTINUE
            mps_regData->e_smState = CONTINUE;
            break;
        case M2:
            if(WSC_ID_MESSAGE_M3 != msgType)
            {
                TUTRACE((TUTRACE_ERR, "REGSM: Expected M3, received %d\n",
                          msgType));
                return;
            }

            err = mpc_regProt->ProcessMessageM3(mps_regData, msg);
			if(WSC_SUCCESS != err)
            {
                stringPrintf(errMsg, 256, "ProcessMessageM3: %d", err);
                //Send a NACK to the Enrollee
				SendNack(WSC_MESSAGE_PROCESSING_ERROR);
                return;//for vista //throw errMsg;
            }
            mps_regData->e_lastMsgRecd = M3;

            err = mpc_regProt->BuildMessageM4(mps_regData, tempBuf);
            if(WSC_SUCCESS != err)
            {
                stringPrintf(errMsg, 256, "BuildMessageM4: %d", err);
                //Send a NACK to the Enrollee
				SendNack(WSC_MESSAGE_PROCESSING_ERROR);
                throw errMsg;
            }
            mps_regData->e_lastMsgSent = M4;
            
            outBuf.Append(tempBuf.Length(), tempBuf.GetBuf());

            //Now send the message to the transport
            err = mpc_trans->TrWrite(m_transportType, 
                               (char *)outBuf.GetBuf(), 
                               outBuf.Length());
            if(WSC_SUCCESS != err)
            {
                mps_regData->e_smState = FAILURE;
                TUTRACE((TUTRACE_ERR, "REGSM: TrWrite generated an "
                                      "error: %d\n", err));
                return;
            }

            //set the message state to CONTINUE
            mps_regData->e_smState = CONTINUE;
            break;
        case M2D:
            if(WSC_ID_MESSAGE_ACK == msgType)
            {
                err = mpc_regProt->ProcessMessageAck(mps_regData, msg);
				if ( WSC_SUCCESS == err && g_TotalUPnPEventSubscribers == 0 ) {
					stringPrintf(errMsg, 256, "No external registrars, send EAP_FAIL and restart", err);
					throw errMsg;
				}
            }
            else if(WSC_ID_MESSAGE_NACK == msgType)
            {
                uint16 configError;
                err = mpc_regProt->ProcessMessageNack(mps_regData, 
                                                      msg, 
                                                      &configError);
                if(configError != WSC_ERROR_NO_ERROR)
                {
                    TUTRACE((TUTRACE_ERR, "REGSM: Recvd NACK with config err: "
                                          "%d", configError));
                }
            }
            else
            {
                err = RPROT_ERR_WRONG_MSGTYPE;
            }
            if(WSC_SUCCESS != err)
                throw err;
            else
                mps_regData->e_smState = RESTART; //Done processing for now
            break;
        case M4:
            if(WSC_ID_MESSAGE_M5 != msgType)
            {
                TUTRACE((TUTRACE_ERR, "REGSM: Expected M5 (0x09), "
                                      "received %0x\n", msgType));
                return;
            }

            err = mpc_regProt->ProcessMessageM5(mps_regData, msg);
			if(WSC_SUCCESS != err)
            {
                stringPrintf(errMsg, 256, "ProcessMessageM5: %d", err);
                //Send a NACK to the Enrollee
				if (err == RPROT_ERR_CRYPTO) {
					SendNack(WSC_PASSWORD_AUTH_ERROR);
				} else {
					SendNack(WSC_MESSAGE_PROCESSING_ERROR);
				}
                return;//for vista //throw errMsg;
            }
            mps_regData->e_lastMsgRecd = M5;

            err = mpc_regProt->BuildMessageM6(mps_regData, tempBuf);
            if(WSC_SUCCESS != err)
            {
                stringPrintf(errMsg, 256, "BuildMessageM6: %d", err);
                //Send a NACK to the Enrollee
				SendNack(WSC_MESSAGE_PROCESSING_ERROR);
                throw errMsg;
            }
            mps_regData->e_lastMsgSent = M6;            
            
            outBuf.Append(tempBuf.Length(), tempBuf.GetBuf());

            //Now send the message to the transport
            err = mpc_trans->TrWrite(m_transportType, 
                               (char *)outBuf.GetBuf(), 
                               outBuf.Length());
            if(WSC_SUCCESS != err)
            {
                mps_regData->e_smState = FAILURE;
                TUTRACE((TUTRACE_ERR, "REGSM: TrWrite generated an "
                                      "error: %d\n", err));
                return;
            }

            //set the message state to CONTINUE
            mps_regData->e_smState = CONTINUE;
            break;
        case M6:
            if(WSC_ID_MESSAGE_M7 != msgType)
            {
                TUTRACE((TUTRACE_ERR, "REGSM: Expected M7, received %d\n",
                          msgType));
                return;
            }

            err = mpc_regProt->ProcessMessageM7(mps_regData, msg,&encrSettings);
			if(WSC_SUCCESS != err)
            {
                stringPrintf(errMsg, 256, "ProcessMessageM7: %d", err);
                //Send a NACK to the Enrollee
				if (err == RPROT_ERR_CRYPTO) {
					SendNack(WSC_PASSWORD_AUTH_ERROR);
				} else {
					SendNack(WSC_MESSAGE_PROCESSING_ERROR);
				}
                return;//for vista //throw errMsg;
            }
            mps_regData->e_lastMsgRecd = M7;
            mp_peerEncrSettings = encrSettings;

            //Build message 8 with the appropriate encrypted settings
            if(mps_regData->p_enrolleeInfo->b_ap)
            {
                err = mpc_regProt->BuildMessageM8(mps_regData, 
                                                  tempBuf, 
                                                  mps_regData->apEncrSettings);
            }
            else
            {
                err = mpc_regProt->BuildMessageM8(mps_regData, 
                                                  tempBuf, 
                                                  mps_regData->staEncrSettings);
            }
            if(WSC_SUCCESS != err)
            {
                stringPrintf(errMsg, 256, "BuildMessageM8: %d", err);
                //Send a NACK to the Enrollee
				SendNack(WSC_MESSAGE_PROCESSING_ERROR);
                throw errMsg;
            }
            mps_regData->e_lastMsgSent = M8;

            outBuf.Append(tempBuf.Length(), tempBuf.GetBuf());

            //Now send the message to the transport
            err = mpc_trans->TrWrite(m_transportType, 
                               (char *)outBuf.GetBuf(), 
                               outBuf.Length());
            if(WSC_SUCCESS != err)
            {
                mps_regData->e_smState = FAILURE;
                TUTRACE((TUTRACE_ERR, "REGSM: TrWrite generated an "
                                      "error: %d\n", err));
                return;
            }

            //set the message state to continue
            mps_regData->e_smState = CONTINUE;
            break;
        case M8:
            if(WSC_ID_MESSAGE_DONE != msgType)
            {
                TUTRACE((TUTRACE_ERR, "REGSM: Expected DONE, received %d\n",
                          msgType));
                return;
            }

            err = mpc_regProt->ProcessMessageDone(mps_regData, msg);
			if (WSC_SUCCESS != err)
            {
                stringPrintf(errMsg, 256, "ProcessMessageDone: %d", err);
                return;//for vista //throw errMsg;
            }
            
            if(	mps_regData->p_enrolleeInfo->b_ap && 
				(m_transportType == TRANSPORT_TYPE_EAP)) {
				// When setting up an AP over EAP, the Done msg from the AP comes in EAP-Request, 
				// so send ACK in response.
                SendAck();  
			}

            //set the message state to success
            mps_regData->e_smState = SUCCESS;
			
            WscLock(mp_ptimerLock);
            m_ptimerStatus = 1; 
            WscUnlock(mp_ptimerLock);
            break;
        default:
            if(WSC_ID_MESSAGE_NACK == msgType)
            {
                uint16 configError;
                err = mpc_regProt->ProcessMessageNack(mps_regData,
                                                      msg,
                                                      &configError);
				if (WSC_SUCCESS != err)
                {
                    stringPrintf(errMsg, 256, "ProcessMessageNack: %d", err);
                    throw errMsg;
                }
                TUTRACE((TUTRACE_ERR, "REGSM: Recvd NACK with err code: %d", 
                                       configError));
                mps_regData->e_smState = FAILURE;
            }
            else
            {
                throw "Unexpected message received";
            }
        }
    }
    catch(uint32 err)
    {
        TUTRACE((TUTRACE_ERR, "REGSM: HandleMessage threw an exception: %d\n",
                              err));
        //send an empty message to the transport
        // mpc_trans->TrWrite(m_transportType, NULL, 0); // this will be done by RegSMStep

        //set the message state to failure
        mps_regData->e_smState = FAILURE;

    }
    catch(char *str)
    {
        TUTRACE((TUTRACE_ERR, "REGSM: HandleMessage threw an exception: %s\n",
                              str));
        //send an empty message to the transport
        // mpc_trans->TrWrite(m_transportType, NULL, 0); // this will be done by RegSMStep

        //set the message state to failure
        mps_regData->e_smState = FAILURE;
    }
    catch(...)
    {
        TUTRACE((TUTRACE_ERR, "REGSM: HandleMessage threw an unknown "
                              "exception\n"));
        //send an empty message to the transport
        // mpc_trans->TrWrite(m_transportType, NULL, 0); // this will be done by RegSMStep

        //set the message state to failure
        mps_regData->e_smState = FAILURE;
    }
}

//added for wsc timeout

void *
CRegistrarSM::ProtocolTimerThread(IN void *p_data)
{
    TUTRACE((TUTRACE_INFO, "REGSM: In ProtocolTimerThread\n"));
    CRegistrarSM *registrar = (CRegistrarSM *)p_data;
    //int32 timercount = 0;
    int32 messagetimeoutcount=0 ;
    int32 templastMsgSent = 0;
    //We need to wait 
    while ((messagetimeoutcount++ < WSC_MESSAGE_TIMEOUT)/* && (timercount++ < WSC_PROTOCOL_TIMEOUT)*/ )
    {
#ifndef __linux__
        Sleep(1*1000);
#else
        WscSleep(1);  
#endif
        if ((templastMsgSent != registrar->mps_regData->e_lastMsgSent) || (M8 == registrar->mps_regData->e_lastMsgSent))
			messagetimeoutcount=0;

        templastMsgSent = registrar->mps_regData->e_lastMsgSent ;
        WscLock(registrar->mp_ptimerLock);
        if(1 == registrar->m_ptimerStatus)
        {
            //Nothing to be done. 
            WscUnlock(registrar->mp_ptimerLock);
		    TUTRACE((TUTRACE_INFO, "REGSM: WSC protocol OK! ProtocolTimer disable\n"));
            return NULL;
        }
        WscUnlock(registrar->mp_ptimerLock);
    }
	
    TUTRACE((TUTRACE_ERR, "REGSM: ProtocolTimerThread message timeout !!\n"));
    return NULL;
}
//add --


